@@ -44,7 +44,6 @@ from oslo_service import periodic_task
 from oslo_utils import excutils
 from oslo_utils import importutils
 from oslo_utils import timeutils
-import six
 
 from cinder.backup import rpcapi as backup_rpcapi
 from cinder import context
@@ -52,6 +51,8 @@ from cinder import exception
 from cinder.i18n import _
 from cinder.keymgr import migration as key_migration
 from cinder import manager
+from cinder.message import api as message_api
+from cinder.message import message_field
 from cinder import objects
 from cinder.objects import fields
 from cinder import quota
@@ -136,6 +137,7 @@ class BackupManager(manager.SchedulerDependentManager):
                         self.driver_name, new_name)
             self.driver_name = new_name
         self.service = importutils.import_class(self.driver_name)
+        self.message_api = message_api.API()
 
     def init_host(self, **kwargs):
         """Run initialization needed for a standalone service."""
@@ -156,7 +158,7 @@ class BackupManager(manager.SchedulerDependentManager):
         self.publish_service_capabilities(ctxt)
 
     def _setup_backup_driver(self, ctxt):
-        backup_service = self.service(context=ctxt, db=self.db)
+        backup_service = self.service(context=ctxt)
         backup_service.check_for_setup_error()
         self.is_initialized = True
         raise loopingcall.LoopingCallDone()
@@ -340,7 +342,9 @@ class BackupManager(manager.SchedulerDependentManager):
         snapshot = objects.Snapshot.get_by_id(
             context, snapshot_id) if snapshot_id else None
         previous_status = volume.get('previous_status', None)
-        updates = {}
+        context.message_resource_id = backup.id
+        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
+        context.message_action = message_field.Action.BACKUP_CREATE
         if snapshot_id:
             log_message = ('Create backup started, backup: %(backup_id)s '
                            'volume: %(volume_id)s snapshot: %(snapshot_id)s.'
@@ -387,17 +391,25 @@ class BackupManager(manager.SchedulerDependentManager):
                 'actual_status': actual_status,
             }
             volume_utils.update_backup_error(backup, err)
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.BACKUP_INVALID_STATE)
             raise exception.InvalidBackup(reason=err)
 
         try:
             if not self.is_working():
                 err = _('Create backup aborted due to backup service is down.')
                 volume_utils.update_backup_error(backup, err)
+                self.message_api.create_from_request_context(
+                    context,
+                    detail=message_field.Detail.BACKUP_SERVICE_DOWN)
                 raise exception.InvalidBackup(reason=err)
 
             backup.service = self.driver_name
             backup.save()
-            updates = self._run_backup(context, backup, volume)
+
+            # Start backup, then continue_backup, then finish_backup
+            self._start_backup(context, backup, volume)
         except Exception as err:
             with excutils.save_and_reraise_exception():
                 if snapshot_id:
@@ -408,7 +420,129 @@ class BackupManager(manager.SchedulerDependentManager):
                         context, volume_id,
                         {'status': previous_status,
                          'previous_status': 'error_backing-up'})
-                volume_utils.update_backup_error(backup, six.text_type(err))
+                volume_utils.update_backup_error(backup, str(err))
+
+    def _start_backup(self, context, backup, volume):
+        """This starts the backup process.
+
+           First we have to get the backup device from the volume manager.
+           This can take a long time to complete.  Once the volume manager
+           is done creating/getting the backup device, then we get a callback
+           to complete the process of backing up the volume.
+
+        """
+        # Save a copy of the encryption key ID in case the volume is deleted.
+        if (volume.encryption_key_id is not None and
+                backup.encryption_key_id is None):
+            backup.encryption_key_id = volume_utils.clone_encryption_key(
+                context,
+                key_manager.API(CONF),
+                volume.encryption_key_id)
+            backup.save()
+
+        # This is an async call to the volume manager.  We will get a
+        # callback from the volume manager to continue once it's done.
+        LOG.info("Call Volume Manager to get_backup_device for %s", backup)
+        self.volume_rpcapi.get_backup_device(context, backup, volume)
+
+    def continue_backup(self, context, backup, backup_device):
+        """This is the callback from the volume manager to continue."""
+        message_created = False
+        volume_id = backup.volume_id
+        volume = objects.Volume.get_by_id(context, volume_id)
+        snapshot_id = backup.snapshot_id
+        snapshot = objects.Snapshot.get_by_id(
+            context, snapshot_id) if snapshot_id else None
+        previous_status = volume.get('previous_status', None)
+
+        backup_service = self.service(context)
+        properties = volume_utils.brick_get_connector_properties()
+
+        updates = {}
+        try:
+            try:
+                attach_info = self._attach_device(context,
+                                                  backup_device.device_obj,
+                                                  properties,
+                                                  backup_device.is_snapshot)
+            except Exception:
+                with excutils.save_and_reraise_exception():
+                    if not message_created:
+                        message_created = True
+                        self.message_api.create_from_request_context(
+                            context,
+                            detail=message_field.Detail.ATTACH_ERROR)
+
+            try:
+                device_path = attach_info['device']['path']
+                if (isinstance(device_path, str) and
+                        not os.path.isdir(device_path)):
+                    if backup_device.secure_enabled:
+                        with open(device_path, 'rb') as device_file:
+                            updates = backup_service.backup(
+                                backup, tpool.Proxy(device_file))
+                    else:
+                        with utils.temporary_chown(device_path):
+                            with open(device_path, 'rb') as device_file:
+                                updates = backup_service.backup(
+                                    backup, tpool.Proxy(device_file))
+                # device_path is already file-like so no need to open it
+                else:
+                    updates = backup_service.backup(backup,
+                                                    tpool.Proxy(device_path))
+            except Exception:
+                with excutils.save_and_reraise_exception():
+                    if not message_created:
+                        message_created = True
+                        self.message_api.create_from_request_context(
+                            context,
+                            detail=
+                            message_field.Detail.BACKUP_CREATE_DRIVER_ERROR)
+            finally:
+                try:
+                    self._detach_device(context, attach_info,
+                                        backup_device.device_obj, properties,
+                                        backup_device.is_snapshot, force=True,
+                                        ignore_errors=True)
+                except Exception:
+                    with excutils.save_and_reraise_exception():
+                        if not message_created:
+                            message_created = True
+                            self.message_api.create_from_request_context(
+                                context,
+                                detail=
+                                message_field.Detail.DETACH_ERROR)
+        except Exception as err:
+            with excutils.save_and_reraise_exception():
+                if snapshot_id:
+                    snapshot.status = fields.SnapshotStatus.AVAILABLE
+                    snapshot.save()
+                else:
+                    self.db.volume_update(
+                        context, volume_id,
+                        {'status': previous_status,
+                         'previous_status': 'error_backing-up'})
+                volume_utils.update_backup_error(backup, str(err))
+        finally:
+            with backup.as_read_deleted():
+                backup.refresh()
+            try:
+                self._cleanup_temp_volumes_snapshots_when_backup_created(
+                    context, backup)
+            except Exception:
+                with excutils.save_and_reraise_exception():
+                    if not message_created:
+                        self.message_api.create_from_request_context(
+                            context,
+                            detail=
+                            message_field.Detail.BACKUP_CREATE_CLEANUP_ERROR)
+
+        self._finish_backup(context, backup, volume, updates)
+
+    def _finish_backup(self, context, backup, volume, updates):
+        volume_id = backup.volume_id
+        snapshot_id = backup.snapshot_id
+        previous_status = volume.get('previous_status', None)
 
         # Restore the original status.
         if snapshot_id:
@@ -444,64 +578,9 @@ class BackupManager(manager.SchedulerDependentManager):
         LOG.info('Create backup %s. backup: %s.', completion_msg, backup.id)
         self._notify_about_backup_usage(context, backup, "create.end")
 
-    def _run_backup(self, context, backup, volume):
-        # Save a copy of the encryption key ID in case the volume is deleted.
-        if (volume.encryption_key_id is not None and
-                backup.encryption_key_id is None):
-            backup.encryption_key_id = volume_utils.clone_encryption_key(
-                context,
-                key_manager.API(CONF),
-                volume.encryption_key_id)
-            backup.save()
-
-        backup_service = self.service(context)
-
-        properties = utils.brick_get_connector_properties()
-
-        # NOTE(geguileo): Not all I/O disk operations properly do greenthread
-        # context switching and may end up blocking the greenthread, so we go
-        # with native threads proxy-wrapping the device file object.
-        try:
-            backup_device = self.volume_rpcapi.get_backup_device(context,
-                                                                 backup,
-                                                                 volume)
-            attach_info = self._attach_device(context,
-                                              backup_device.device_obj,
-                                              properties,
-                                              backup_device.is_snapshot)
-            try:
-                device_path = attach_info['device']['path']
-                if (isinstance(device_path, six.string_types) and
-                        not os.path.isdir(device_path)):
-                    if backup_device.secure_enabled:
-                        with open(device_path, 'rb') as device_file:
-                            updates = backup_service.backup(
-                                backup, tpool.Proxy(device_file))
-                    else:
-                        with utils.temporary_chown(device_path):
-                            with open(device_path, 'rb') as device_file:
-                                updates = backup_service.backup(
-                                    backup, tpool.Proxy(device_file))
-                # device_path is already file-like so no need to open it
-                else:
-                    updates = backup_service.backup(backup,
-                                                    tpool.Proxy(device_path))
-
-            finally:
-                self._detach_device(context, attach_info,
-                                    backup_device.device_obj, properties,
-                                    backup_device.is_snapshot, force=True,
-                                    ignore_errors=True)
-        finally:
-            with backup.as_read_deleted():
-                backup.refresh()
-            self._cleanup_temp_volumes_snapshots_when_backup_created(
-                context, backup)
-        return updates
-
     def _is_our_backup(self, backup):
         # Accept strings and Service OVO
-        if not isinstance(backup, six.string_types):
+        if not isinstance(backup, str):
             backup = backup.service
 
         if not backup:
@@ -524,6 +603,9 @@ class BackupManager(manager.SchedulerDependentManager):
     @utils.limit_operations
     def restore_backup(self, context, backup, volume_id):
         """Restore volume backups from configured backup service."""
+        context.message_resource_id = backup.id
+        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
+        context.message_action = message_field.Action.BACKUP_RESTORE
         LOG.info('Restore backup started, backup: %(backup_id)s '
                  'volume: %(volume_id)s.',
                  {'backup_id': backup.id, 'volume_id': volume_id})
@@ -547,6 +629,12 @@ class BackupManager(manager.SchedulerDependentManager):
                  (fields.VolumeStatus.ERROR if
                   volume_previous_status == fields.VolumeStatus.CREATING else
                   fields.VolumeStatus.ERROR_RESTORING)})
+            self.message_api.create(
+                context,
+                action=message_field.Action.BACKUP_RESTORE,
+                resource_type=message_field.Resource.VOLUME_BACKUP,
+                resource_uuid=volume.id,
+                detail=message_field.Detail.VOLUME_INVALID_STATE)
             raise exception.InvalidVolume(reason=err)
 
         expected_status = fields.BackupStatus.RESTORING
@@ -559,6 +647,9 @@ class BackupManager(manager.SchedulerDependentManager):
             volume_utils.update_backup_error(backup, err)
             self.db.volume_update(context, volume_id,
                                   {'status': fields.VolumeStatus.ERROR})
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.BACKUP_INVALID_STATE)
             raise exception.InvalidBackup(reason=err)
 
         if volume['size'] > backup['size']:
@@ -629,14 +720,21 @@ class BackupManager(manager.SchedulerDependentManager):
         self._notify_about_backup_usage(context, backup, "restore.end")
 
     def _run_restore(self, context, backup, volume):
+        message_created = False
         orig_key_id = volume.encryption_key_id
         backup_service = self.service(context)
 
-        properties = utils.brick_get_connector_properties()
+        properties = volume_utils.brick_get_connector_properties()
         secure_enabled = (
             self.volume_rpcapi.secure_file_operations_enabled(context,
                                                               volume))
-        attach_info = self._attach_device(context, volume, properties)
+        try:
+            attach_info = self._attach_device(context, volume, properties)
+        except Exception:
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.ATTACH_ERROR)
+            raise
 
         # NOTE(geguileo): Not all I/O disk operations properly do greenthread
         # context switching and may end up blocking the greenthread, so we go
@@ -644,7 +742,7 @@ class BackupManager(manager.SchedulerDependentManager):
         try:
             device_path = attach_info['device']['path']
             open_mode = 'rb+' if os.name == 'nt' else 'wb'
-            if (isinstance(device_path, six.string_types) and
+            if (isinstance(device_path, str) and
                     not os.path.isdir(device_path)):
                 if secure_enabled:
                     with open(device_path, open_mode) as device_file:
@@ -665,10 +763,25 @@ class BackupManager(manager.SchedulerDependentManager):
             LOG.exception('Restoring backup %(backup_id)s to volume '
                           '%(volume_id)s failed.', {'backup_id': backup.id,
                                                     'volume_id': volume.id})
+            # We set message_create to True before creating the
+            # message because if the message create call fails
+            # and is catched by the base/outer exception handler
+            # then we will end up storing a wrong message
+            message_created = True
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.BACKUP_RESTORE_ERROR)
             raise
         finally:
-            self._detach_device(context, attach_info, volume, properties,
-                                force=True)
+            try:
+                self._detach_device(context, attach_info, volume, properties,
+                                    force=True)
+            except Exception:
+                if not message_created:
+                    self.message_api.create_from_request_context(
+                        context,
+                        detail=message_field.Detail.DETACH_ERROR)
+                raise
 
         # Regardless of whether the restore was successful, do some
         # housekeeping to ensure the restored volume's encryption key ID is
@@ -718,6 +831,9 @@ class BackupManager(manager.SchedulerDependentManager):
 
         self._notify_about_backup_usage(context, backup, "delete.start")
 
+        context.message_resource_id = backup.id
+        context.message_resource_type = message_field.Resource.VOLUME_BACKUP
+        context.message_action = message_field.Action.BACKUP_DELETE
         expected_status = fields.BackupStatus.DELETING
         actual_status = backup.status
         if actual_status != expected_status:
@@ -726,12 +842,18 @@ class BackupManager(manager.SchedulerDependentManager):
                 % {'expected_status': expected_status,
                    'actual_status': actual_status}
             volume_utils.update_backup_error(backup, err)
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.BACKUP_INVALID_STATE)
             raise exception.InvalidBackup(reason=err)
 
         if backup.service and not self.is_working():
             err = _('Delete backup is aborted due to backup service is down.')
             status = fields.BackupStatus.ERROR_DELETING
             volume_utils.update_backup_error(backup, err, status)
+            self.message_api.create_from_request_context(
+                context,
+                detail=message_field.Detail.BACKUP_SERVICE_DOWN)
             raise exception.InvalidBackup(reason=err)
 
         if not self._is_our_backup(backup):
@@ -750,8 +872,10 @@ class BackupManager(manager.SchedulerDependentManager):
                 backup_service.delete_backup(backup)
             except Exception as err:
                 with excutils.save_and_reraise_exception():
-                    volume_utils.update_backup_error(backup,
-                                                     six.text_type(err))
+                    volume_utils.update_backup_error(backup, str(err))
+                    self.message_api.create_from_request_context(
+                        context,
+                        detail=message_field.Detail.BACKUP_DELETE_DRIVER_ERROR)
 
         # Get reservations
         try:
@@ -841,7 +965,7 @@ class BackupManager(manager.SchedulerDependentManager):
             backup_url = backup.encode_record(driver_info=driver_info)
             backup_record['backup_url'] = backup_url
         except Exception as err:
-            msg = six.text_type(err)
+            msg = str(err)
             raise exception.InvalidBackup(reason=msg)
 
         LOG.info('Export record finished, backup %s exported.', backup.id)
@@ -895,7 +1019,7 @@ class BackupManager(manager.SchedulerDependentManager):
                 backup_service = self.service(context)
                 backup_service.import_record(backup, driver_options)
             except Exception as err:
-                msg = six.text_type(err)
+                msg = str(err)
                 volume_utils.update_backup_error(backup, msg)
                 raise exception.InvalidBackup(reason=msg)
 
@@ -1023,7 +1147,7 @@ class BackupManager(manager.SchedulerDependentManager):
                     LOG.warning("Failed to terminate the connection "
                                 "of volume %(volume_id)s, but it is "
                                 "acceptable.",
-                                {'volume_id', volume.id})
+                                {'volume_id': volume.id})
 
     def _attach_snapshot(self, ctxt, snapshot, properties):
         """Attach a snapshot."""
@@ -1041,14 +1165,14 @@ class BackupManager(manager.SchedulerDependentManager):
                     LOG.warning("Failed to terminate the connection "
                                 "of snapshot %(snapshot_id)s, but it is "
                                 "acceptable.",
-                                {'snapshot_id', snapshot.id})
+                                {'snapshot_id': snapshot.id})
 
     def _connect_device(self, conn):
         """Establish connection to device."""
         use_multipath = CONF.use_multipath_for_image_xfer
         device_scan_attempts = CONF.num_volume_device_scan_tries
         protocol = conn['driver_volume_type']
-        connector = utils.brick_get_connector(
+        connector = volume_utils.brick_get_connector(
             protocol,
             use_multipath=use_multipath,
             device_scan_attempts=device_scan_attempts,
@@ -1074,7 +1198,7 @@ class BackupManager(manager.SchedulerDependentManager):
         else:
             rpcapi.terminate_connection_snapshot(ctxt, device,
                                                  properties, force=force)
-            rpcapi.remove_export_snapshot(ctxt, device)
+            rpcapi.remove_export_snapshot(ctxt, device, sync=True)
 
     def is_working(self):
         return self.is_initialized
@@ -1087,8 +1211,6 @@ class BackupManager(manager.SchedulerDependentManager):
         self._publish_service_capabilities(context)
 
     def _report_driver_status(self, context):
-        if not self.is_working():
-            self.setup_backup_backend(context)
         backup_stats = {
             'backend_state': self.is_working(),
             'driver_name': self.driver_name,
